package de.cau.cs.kieler.tutorials.fdl22.turingmachinevis

import javax.inject.Inject

import de.cau.cs.kieler.klighd.kgraph.KNode
import de.cau.cs.kieler.klighd.krendering.KRenderingFactory
import de.cau.cs.kieler.klighd.krendering.extensions.KNodeExtensions
import de.cau.cs.kieler.klighd.krendering.extensions.KEdgeExtensions
import de.cau.cs.kieler.klighd.krendering.extensions.KPortExtensions
import de.cau.cs.kieler.klighd.krendering.extensions.KLabelExtensions
import de.cau.cs.kieler.klighd.krendering.extensions.KRenderingExtensions
import de.cau.cs.kieler.klighd.krendering.extensions.KContainerRenderingExtensions
import de.cau.cs.kieler.klighd.krendering.extensions.KPolylineExtensions
import de.cau.cs.kieler.klighd.krendering.extensions.KColorExtensions
import de.cau.cs.kieler.klighd.syntheses.AbstractDiagramSynthesis

import static de.cau.cs.kieler.klighd.syntheses.DiagramLayoutOptions.*

import static extension de.cau.cs.kieler.klighd.syntheses.DiagramSyntheses.*

import de.cau.cs.kieler.tutorials.klighd.turingmachine.TuringMachine
import de.cau.cs.kieler.tutorials.klighd.turingmachine.State
import de.cau.cs.kieler.klighd.KlighdConstants
import org.eclipse.elk.core.options.CoreOptions
import java.util.EnumSet
import org.eclipse.elk.core.options.SizeConstraint
import de.cau.cs.kieler.klighd.kgraph.KEdge
import de.cau.cs.kieler.tutorials.klighd.turingmachine.Transition
import de.cau.cs.kieler.klighd.kgraph.util.KGraphUtil
import org.eclipse.elk.core.options.Direction
import org.eclipse.elk.alg.layered.options.LayeredOptions
import org.eclipse.elk.alg.layered.options.OrderingStrategy
import org.eclipse.elk.core.options.EdgeRouting
import org.eclipse.elk.alg.force.options.ForceOptions
import de.cau.cs.kieler.klighd.actions.CollapseExpandAction
import de.cau.cs.kieler.klighd.SynthesisOption
import de.cau.cs.kieler.klighd.krendering.Colors

class TuringMachineDiagramSynthesis extends AbstractDiagramSynthesis<TuringMachine> {
    
    @Inject extension KNodeExtensions
    @Inject extension KEdgeExtensions
    @Inject extension KPortExtensions
    @Inject extension KLabelExtensions
    @Inject extension KRenderingExtensions
    @Inject extension KContainerRenderingExtensions
    @Inject extension KPolylineExtensions
    @Inject extension KColorExtensions
    extension KRenderingFactory = KRenderingFactory.eINSTANCE
    
    override KNode transform(TuringMachine turingMachine) {
        val root = turingMachine.createNode().associateWith(turingMachine)
        
        // Configure Layout
        root.setLayoutOption(CoreOptions.DIRECTION, Direction.RIGHT)
        root.setLayoutOption(LayeredOptions.CONSIDER_MODEL_ORDER_STRATEGY, OrderingStrategy.NODES_AND_EDGES)
        root.setLayoutOption(CoreOptions.ALGORITHM, LayeredOptions.ALGORITHM_ID)
//	    root.setLayoutOption(CoreOptions.ALGORITHM, ForceOptions.ALGORITHM_ID)
        root.setLayoutOption(CoreOptions.EDGE_ROUTING, EdgeRouting.SPLINES)
        
        for (state : turingMachine.states) {
        	root.children += transform(state)
        }
        
        if (initialState !== null) {
        	root.children += transformInitial
        }
        
        return root
    }
	
	def transformInitial() {
		val initialViewNode = initialState.createNode("Start").associateWith(initialState)
		initialViewNode.addText("Start")
		
		val initialViewEdge = initialState.createEdge.associateWith(initialState)
		initialViewEdge.addSpline(2).addHeadArrowDecorator
		
		initialViewNode.outgoingEdges += initialViewEdge
		initialViewEdge.target = initialState.transform
		
		return initialViewNode
	}
    
    private def create viewNode : state.createNode() transform(State state) {
    	viewNode.associateWith(state)
    	
    	if (state.final) {
    		// Collapsed View
    		val ellipse = viewNode.addEllipse
    			ellipse.setAsCollapsedView
    			ellipse.addDoubleClickAction(CollapseExpandAction.ID)
    			ellipse.lineWidth = 2
    			ellipse.tooltip = "This is the final state."
    			val innerEllipse = ellipse.addEllipse
    				innerEllipse.lineWidth = 2
    				innerEllipse.setAreaPlacementData.from(LEFT, 3, 0, TOP, 3, 0).to(RIGHT, 3, 0, BOTTOM, 3, 0)
    				
    				val text = innerEllipse.addText(state.name)
		    			text.setAreaPlacementData.from(LEFT, 10, 0, TOP, 8, 0).to(RIGHT, 10, 0, BOTTOM, 8, 0)
			
			
    		// Expanded View
    		val eellipse = viewNode.addEllipse
    			eellipse.setAsExpandedView
    			eellipse.addDoubleClickAction(CollapseExpandAction.ID)
    			eellipse.lineWidth = 2
    			val einnerEllipse = eellipse.addEllipse
    				einnerEllipse.lineWidth = 2
    				einnerEllipse.setAreaPlacementData.from(LEFT, 3, 0, TOP, 3, 0).to(RIGHT, 3, 0, BOTTOM, 3, 0)
    				einnerEllipse.gridPlacement = 1
    				val etext = einnerEllipse.addText(state.name)
		    			etext.setGridPlacementData.from(LEFT, 10, 0, TOP, 8, 0).to(RIGHT, 10, 0, BOTTOM, 8, 0)
    				val etext2 = einnerEllipse.addText("This is the final state.")
		    			etext2.setGridPlacementData.from(LEFT, 20, 0, TOP, 16, 0).to(RIGHT, 20, 0, BOTTOM, 16, 0)
    	} else {
	    	val rectangle = viewNode.addRoundedRectangle(4, 4, 2)
	    		if (state.initial) {
	    			initialState = state
	    			rectangle.background = Colors.AQUAMARINE
	    		}
	    	
	    		val text = rectangle.addText(state.name)
	    			text.setAreaPlacementData.from(LEFT, 10, 0, TOP, 8, 0).to(RIGHT, 10, 0, BOTTOM, 8, 0)
    	}
    	
    	for (transition : state.outgoingTransitions) {
    		viewNode.outgoingEdges += transform(transition)
    	}
    }
    
    private def KEdge transform(Transition transition) {
    	val viewEdge = transition.createEdge.associateWith(transition)
    	
//    	viewEdge.addPolyline(2).addHeadArrowDecorator
    	viewEdge.addSpline(2).addHeadArrowDecorator
    	
    	if (usedContext.getOptionValue(TRANSITION_LABELS) === true) {
	    	// Add a label to the edge
	    	val label = viewEdge.createLabel.associateWith(transition)
	    	val labelText = transition.trigger + " / "
	    		+ transition.action + " "
	    		+ transition.newChar + ", "
	    		+ transition.direction
			label.configureCenterEdgeLabel(labelText)
    	}
    	
    	viewEdge.target = transform(transition.targetState)
    	
    	return viewEdge
    }
    
    public static final SynthesisOption TRANSITION_LABELS = SynthesisOption.createCheckOption("Show Transition Labels", true)
    	.description = "Turns the labels for the transitions on or off."
	
	State initialState
    
    
    override getDisplayedSynthesisOptions() {
    	return #[
    		TRANSITION_LABELS
    	]
    }
}
